#define _GNU_SOURCE
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <dlfcn.h>

#include "module/module_api.h"
#include "def.h"
#include "utils.h" /* for fg_colors enumeration */

void color_printf(int level, enum fg_colors color, char *format, ...) {
	fprintf(stderr, "\033[%d;40m", color);

	va_list args;
	va_start (args, format);

	char *str;
	vasprintf(&str, format, args);
#ifndef VERBOSE_SYSLOG
	if (level != LOG_DEBUG && level != LOG_INFO) {
#endif
		openlog("katrin", LOG_CONS | LOG_PID, LOG_DAEMON);
		syslog(level, "%s", str);
#ifndef VERBOSE_SYSLOG
	}
#endif
	va_end (args);
	fputs(str, stderr);
	free(str);
	fputs("\033[0m\n", stderr);
}

char *get_param_value(const char *pname, struct paramlist *pl) {
//	debug("get_param_value");

	for ( ; pl != NULL; pl = pl->next) {
		if (strcmp(pname, pl->param) == 0) {
			return pl->value;
		}
	}

	debug("return NULL from get_param_value");
	return NULL;
}

int add_param(struct paramlist **top, const char *param, const char *value) {
	struct paramlist *p;
	debug("add_param(param=%s, value=%s)", param, value);
	p = (struct paramlist *)malloc(sizeof(struct paramlist));
	if (p == NULL) {
		err("malloc: can't allocate memory!");
		exit(EXIT_FAILURE);
	}
	strcpy(p->param, param);
	strcpy(p->value, value);
	debug("top of paramlist %d",*top);
	p->next = *top;
	*top = p;

	debug("param added");
	return 0;
}

int add_intparam(struct paramlist **top, const char *param, int value) {
	char v[MAX_PARAM_VALUE_LENGTH];
	sprintf(v, "%d", value);
	return add_param(top, param, v);
}

void paramlist2strlists(struct paramlist *params, char *field_list, char *value_list, char *middle_str_f, char *end_str_f, char *middle_str_v, char *end_str_v) {
	debug("paramlist2strlists");
	int first = 1;
	struct paramlist *p = NULL;
	field_list[0] = '\0';
	value_list[0] = '\0';
	for (p = params; p!=NULL; p = p->next) {
		if (first==1) {
				strcat(field_list, end_str_f);
				strcat(value_list, end_str_v);
				first=0;
		}
		else {
				strcat(field_list, middle_str_f);
				strcat(value_list, middle_str_v);
		}
		strcat(field_list, p->param);
		strcat(field_list, end_str_f);

		debug("p->param = %s; p->value = %s", p->param, p->value);
		strcat(value_list, p->value);
		strcat(value_list, end_str_v);
	}
	debug("exit from paramlist2strlists");
}

/* Function converts paramlist to keyvalue list like this:
 *  'param1=value1' 'param2=value2' ...
 * It is very useful for pass params to shell script
 */
void paramlist2keyvaluelist(struct paramlist *params, char *kvlist) {
	debug("paramlist2keyvaluelist");
	struct paramlist *p = NULL;
	kvlist[0] = '\0';
	for (p = params; p!=NULL; p = p->next) {
		strcat(kvlist, " '");
		strncat(kvlist, p->param, MAX_PARAM_NAME_LENGTH);
		strcat(kvlist, "=");
		strcat(kvlist, p->value);
		strcat(kvlist, "'");
		debug("p->param = %s; p->value = %s", p->param, p->value);
	}
	debug("exit from paramlist2keyvaluelist");
}

inline int anulltoi(const char *str) {
	if (str)
		return atoi(str);
	return 0;
}

/* aka constructors */
struct tariff *Tariff(int id, char *week_days) {
	debug("Creating Tariff with id %d",id);
	struct tariff *this = NULL;
	this = (struct tariff *) calloc(1,sizeof(struct tariff));
	this->id = id;
	assert(strlen(week_days) < sizeof(this->week_days));
	strcpy(this->week_days, week_days);
	return this;
}

struct account *Account(int id, double balance, double credit, int blocked, enum accountType type, int client_id, int group_id) {
	debug("Creating Account with id %d, client_id %d, group_id %d",id, client_id, group_id);
	struct account *this = NULL;
	this = (struct account *) calloc(1,sizeof(struct account));
	this->id = id;
	this->balance = balance;
	this->credit = credit;
	this->blocked = blocked;
	this->type = type;
	this->client_id = client_id;
	this->group_id = group_id;
	assert((type == Personal && client_id) || (type == Group && group_id == 1));
	return this;
}

union abonent *Abonent(int client_id, int group_id, struct clientslist *members, enum abonentType type) {
	debug("Creating Abonent with type %d, client_id %d, group_id %d.",type, client_id, group_id);
	union abonent *this = NULL;
	this = (union abonent *) malloc(sizeof(union abonent));
	switch (type) {
		case Client:
			this->client.id = client_id;
			this->client.services=NULL;
			this->client.uparams=NULL;
			break;
		case ClientGroup:
			this->clientgroup.id = group_id; 
			this->clientgroup.members = members; 
			this->clientgroup.gparams=NULL; 
			break;
		default:	err("Unknown abonent type %d",type); return NULL;
	}
	return this;
}

/* aka destructors */
void free_serviceslist(struct serviceslist *sl) {
	struct serviceslist *item = NULL;
	struct serviceslist *p = sl;
	while (p!=NULL) {
		item = p;
		p = p->next;
		free(item);
	}
}


void free_paramlist(struct paramlist *pl) {
	struct paramlist *item = NULL;
	struct paramlist *p = pl;

	while (p!=NULL) {
		item = p;
		p = p->next;
		free(item);
	}
}

void free_abonentslist(struct abonentslist *al) {
	struct abonentslist *next = NULL;
	debug("free_abonents_list %d", al);

	while (al!=NULL) {
		next = al->next;
		free_abonent(al->abonent, al->type);
		free(al);
		al = next;
	}
}

void free_abonent(union abonent *ab, enum abonentType type) {
	debug("free_abonent");
	free_abonent_ptrs(ab, type);
	free(ab);
}

void free_abonent_ptrs(union abonent *ab, enum abonentType type) {
		if (type == Client) 
			free_client_ptrs(&ab->client);
		else 
			free_clientgroup_ptrs(&ab->clientgroup);
}

void free_clientgroup(struct clientgroup *clientgroup) {
	free_clientgroup_ptrs(clientgroup);
	free(clientgroup);
}

void free_clientgroup_ptrs(struct clientgroup *clientgroup) {
	debug("free_clientgroup_ptrs");
	free_clientslist(clientgroup->members);
	free_paramlist(clientgroup->gparams);
	debug("exit from free_clientgroup_ptrs");
}

void free_client(struct client *client) {
	free_client_ptrs(client);
	free(client);
}

void free_client_ptrs(struct client *client) {
	debug("free_client_ptrs");
	free_serviceslist(client->services);
	free_paramlist(client->uparams);
	debug("exit from free_client_ptrs");
}

void free_clientslist(struct clientslist *ul) {
	struct clientslist *item = NULL;
	struct clientslist *p = ul;

	while (p!=NULL) {
		item = p;
		p = p->next;
		free_client(item->client);
		free(item);
	}   
}

/*
 * Build path to library and try to load it.
 * If library not loaded then returns 1.
 * If library successfully loaded then returns 0 and save library
 * hanlde in second argument.
 *
 * NOTE: after working with library handle you always should call
 *	   dlclose() for it!
 *
 **/
int load_library(const char *libtype, const char *libname, void **libhandle) {
	int ret;
	char *path_to_lib;

	ret = asprintf(&path_to_lib, "%s/libkatrin-%s-%s.so", KATRIN_LIB_DIR, libtype, libname);
	if (ret == -1) {
		err("asprintf: can't create path for %s-%s library!", libtype, libname);
		return 1;
	}

	debug("Load library: %s", path_to_lib);

	(*libhandle) = dlopen(path_to_lib, RTLD_LAZY);
	free(path_to_lib);
	if ((*libhandle) == NULL) {
		fprintf(stderr, "%s\n", dlerror());
		return 1;
	}

	return 0;
}


